//-*- c++ -*-
extern "C"{
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
}

#include <iostream.h>
#include <mpi.h>

#define MY_TAG 1

#ifdef DEBUG
int Debug = DEBUG;
#else
int Debug= 0;
#endif

double starttime;
int p;
typedef int ProcessorNo;
ProcessorNo MyProcId;
int Problemsize;
char *progname;

void init(int argc, char* argv[]){
  MPI_Init(&argc,&argv); 
  progname = argv[0];
  starttime = MPI_Wtime();
  MPI_Comm_size(MPI_COMM_WORLD,&p);
  MPI_Comm_rank(MPI_COMM_WORLD,&MyProcId);
  Problemsize = atoi(argv[1]);
  MPI_Bcast(&Problemsize,1,MPI_INT,0,MPI_COMM_WORLD);}

void finish(){
  if (MyProcId == 0) 
    cout << progname << " Problemsize = " << Problemsize << ", Processors = " << p << ", Time = "
         << MPI_Wtime()-starttime << endl << flush;
  int dummy = MPI_Finalize();}

template <class T>
void communicate (ProcessorNo to, ProcessorNo from, T* buf, int size){ 
   MPI_Status stat;  MPI_Request req;
   if (Debug) {printf("sending message of size %d from %d to %d\n",
                       size,MyProcId,to); fflush(stdout);}
   MPI_Isend(buf, size, MPI_BYTE, to, MY_TAG, MPI_COMM_WORLD,&req);
   if (Debug) {printf("%d waiting for message of size %d from %d\n",
                      MyProcId,size,from); fflush(stdout);}
   MPI_Recv(buf, size, MPI_BYTE, from, MPI_ANY_TAG, MPI_COMM_WORLD, &stat); 
   if (Debug) {printf("received message of size %d with tag %d from %d by %d\n",
                      size,stat.MPI_TAG,from,MyProcId); 
     fflush(stdout);}
   MPI_Wait(&req,&stat);}

template <class T>
void communicateMat(ProcessorNo to, ProcessorNo from, T** buf, int rows, int cols){ 
   MPI_Status stat;  MPI_Request req;
   int i, j;
   T* a = new T[rows*cols];
   for(i=0; i<rows; i++)
     for(j=0; j<cols; j++) a[i*cols+j] = buf[i][j];
   int size = rows*cols*sizeof(T);
   if (Debug) {printf("sending message of size %d from %d to %d\n",
                       size,MyProcId,to); fflush(stdout);}
   MPI_Isend(a, size, MPI_BYTE, to, MY_TAG, MPI_COMM_WORLD,&req);
   if (Debug) {printf("%d waiting for message of size %d from %d\n",
                      MyProcId,size,from); fflush(stdout);}
   MPI_Recv(a, size, MPI_BYTE, from, MPI_ANY_TAG, MPI_COMM_WORLD, &stat); 
   if (Debug) {printf("received message of size %d with tag %d from %d by %d\n",
                      size,stat.MPI_TAG,from,MyProcId); fflush(stdout);}
   MPI_Wait(&req,&stat);
   for(i=0; i<rows; i++)
     for(j=0; j<cols; j++) buf[i][j] = a[i*cols+j]  ;
   delete[] a;}

template <class T>
void send(ProcessorNo to, T* buf, int size){ 
   if (Debug) {printf("sending message of size %d from %d to %d\n",
                       size*sizeof(T),MyProcId,to); fflush(stdout);}
   MPI_Send(buf, size*sizeof(T), MPI_BYTE, to, MY_TAG, MPI_COMM_WORLD);}

template <>
void send<double>(ProcessorNo to, double* buf, int size){ 
   if (Debug) {
     cout << "sending message of size " << size << " doubles from " << MyProcId << " to " << to << endl;    
     for(int i=0; i<size; i++)
       cout << *(buf+i) << " ";
     cout << endl;
     fflush(stdout);}
   MPI_Send(buf, size, MPI_DOUBLE, to, MY_TAG, MPI_COMM_WORLD);}

template <class T>
void receive(ProcessorNo from, T* buf, int size){ 
   MPI_Status stat;
   if (Debug) {printf("%d waiting for message of size %d from %d\n",
                      MyProcId,size,from); fflush(stdout);}
   MPI_Recv(buf, size*sizeof(T), MPI_BYTE, from, MPI_ANY_TAG, MPI_COMM_WORLD, &stat); 
   if (Debug) {printf("received message of size %d with tag %d from %d by %d\n",
                      size*sizeof(T),stat.MPI_TAG,from,MyProcId); 
               fflush(stdout);}}

template<> 
void receive<double>(ProcessorNo from, double* buf, int size){ 
   MPI_Status stat;
   if (Debug) {
     cout <<  MyProcId << "waiting for message of size " << size << " doubles from " << from << endl;    
     fflush(stdout);}
   MPI_Recv(buf, size, MPI_DOUBLE, from, MPI_ANY_TAG, MPI_COMM_WORLD, &stat); 
   if (Debug) {
     cout << " received message of size " << size << " doubles with tag";
     cout << stat.MPI_TAG << " from " << from << " by " << MyProcId << endl;    
     for(int i=0; i<size; i++) 
       cout << *(buf+i) << " ";
     cout << endl;
     fflush(stdout);}}

template <class T>
void show(T* A, int size){
#if Output == 1
  T* B = new T[size*p];
  MPI_Gather(A,size*sizeof(T),MPI_BYTE,B,size*sizeof(T),MPI_BYTE,0,MPI_COMM_WORLD);
  if (MyProcId == 0){
    cout << "(";
    for(int i=0; i<size*p; i++) cout << B[i] <<" ";
    cout << ")" << endl << flush;}
  delete[] B;
  if (Debug) {for(int j=0; j<9999999; j++);
  MPI_Barrier(MPI_COMM_WORLD);}
#endif
}

template <class T>
void showRow(T buf[], int cols){
  cout << "(";
  for(int i=0; i<cols; i++) cout << buf[i] << " "; 
  cout << ")" << endl << flush;}

template <class T>
void showMatrix(T* A[], int blockrows, int blockcols, int rows, int cols){ 
#if Output == 1 
  int j, k, br, bc;
  T** B = new T*[blockrows*rows]; 
  for (j=0; j<blockrows*rows; j++) B[j] = new T[blockcols*cols];
  if (MyProcId == 0)
    for (j=0; j<rows; j++)
      for (k=0; k<cols; k++) B[j][k] = A[j][k];
  for (k=1; k<p; k++){
    br = k/blockcols*rows; bc = k%blockcols*cols;
    for (j=0; j<rows; j++)
      if (MyProcId == 0)
        receive(k,&B[br+j][bc],cols); 
      else if (MyProcId == k) 
             for (j=0; j<rows; j++)
               send(0,A[j],cols);}
  if (MyProcId == 0)
    for (j=0; j<blockrows*rows; j++)
      showRow(&B[j][0],blockcols*cols);
  for (j=0; j<blockrows*rows; j++) delete[] B[j];
  delete[] B; 
#endif
}
  
