#include "myheader.h"
#define pi 3.1415926535897932

int log2p;
int n, log2n;
int m;

void complexmult (double r1, double im1, double r2, double im2, double *r, double *im){
  *r = r1*r2 - im1*im2;  *im = r1*im2 +r2*im1;}

void omega(int j, int i, double *r, double *im){
  int b2 = 0; 
  int b = i >> (log2n-j-1); 
  for (int l=0; l<=j; l++){
    b2 = (b & 1) ? 2*b2+1  :  2*b2;
    b >>= 1;}
  double v = 2.0 * pi/n * (b2 << (log2n-j-1));
  *r  = (double) cos(v);
  *im = (double) sin(v);}

int main (int argc, char * argv[]){
  int neighbor;   
  double r, im, r1, im1;  int j, k;

  init(argc,argv); 
  n = Problemsize;
  m = n/p;
  log2n = (int) (log(n)/log(2)+0.1);
  log2p = (int) (log(p)/log(2)+0.1); 
  double* rr = new double[m];
  double* ri = new double[m];
  double* sr = new double[m];
  double* si = new double[m];

  /* init vectors */
  for (int j=0; j<m; j++){ 
    rr[j] = 1.0; ri[j] = 0.0; sr[j] = 0.0; si[j] =0.0;}
  show(rr,m); show(ri,m);
 
  /* global iterations */
  for (k=0; k<log2p; k++){     
    for (j=0; j<m; j++){ 
      sr[j] = rr[j]; si[j] = ri[j];}
    neighbor = MyProcId ^ (1<<(log2p-1-k));
    communicate(neighbor, neighbor, sr, sizeof(double)*m); 
    communicate(neighbor, neighbor, si, sizeof(double)*m);
    for (j=0; j<m; j++){ 
      omega(k,MyProcId*m+j,&r1,&im1);
      if (j & (1 << log2n-1-k)) {
        complexmult(r1,im1,rr[j],ri[j],&r,&im);
        rr[j] = sr[j] + r;   ri[j] = si[j] + im;}
      else {
        complexmult(r1,im1,sr[j],si[j],&r,&im);
      rr[j] += r;    ri[j] += im;}}} 

  /* local iterations */ 
  for (k=log2p; k<log2n; k++){      
    for (j=0; j<m; j++){ 
      sr[j] = rr[j ^ (1<<log2n-1-k)];
      si[j] = ri[j ^ (1<<log2n-1-k)];}
    for (j=0; j<m; j++){
      omega(k,MyProcId*m+j,&r1,&im1);  
      if (j & (1 << log2n-1-k)) {
        complexmult(r1,im1,rr[j],ri[j],&r,&im);
        rr[j] = sr[j] + r;   ri[j] = si[j] + im;}
      else {
        complexmult(r1,im1,sr[j],si[j],&r,&im);
        rr[j] += r;    ri[j] += im;}}}
 
    show(rr,m); show(ri,m); 
    delete[] rr; delete[] ri; delete[] sr; delete[] si; 
    finish();
}
