#include <stdlib.h>
#include "Skeleton.h"
#include <iostream>
#include <fstream>
#include <stdio.h>

#define N 4

int infty =  -1 ^ (1<<31); 

typedef int edgemat_t [N][N];

typedef struct {
  int from, to;
} touredge;

typedef struct {
  touredge tour[N];
  int length;
} tour;

typedef struct {
  edgemat_t edge;
  int lb;
  tour solution;
} problem;

inline ostream& operator<<(ostream& os,  const problem prb){
#if Output == 1
  int sum =  0;
  for(int i=0; i<N; i++)
    for(int j=0; j<N; j++) sum+= prb.edge[i][j];
  os << "problem with lb " << prb.lb << " and edgesum " << sum << endl<< flush;
  return os;
#endif
}

int reduce_row (problem *p, int row)
{
  int col;
  int rowmin;
 
  for (rowmin=infty, col=0; col<N; col++) {
    if (p->edge[row][col] < rowmin) rowmin = p->edge[row][col];
  }
  if (rowmin < infty) {
    for (col=0; col<N; col++) {
      if (p->edge[row][col] < infty)
          p->edge[row][col] -= rowmin;
    }
  }
  return rowmin;
}
 
int reduce_col (problem *p, int col)
{
  int row;
  int colmin;
 
  for (colmin=infty, row=0; row<N; row++) {
    if (p->edge[row][col] < colmin) colmin = p->edge[row][col];
  }
  if (colmin < infty) {
    for (row=0; row<N; row++) {
      if (p->edge[row][col] < infty)
          p->edge[row][col] -= colmin;
    }
  }
  return colmin;
}
 
void reduce_matrix (problem *p)
{
  int row, col, deduct;
 
  for (row = 0; row<N; row++) {
    deduct = reduce_row(p, row);
    if (deduct < infty) p->lb += deduct;
  }
  for (col = 0; col<N; col++) {
    deduct = reduce_col(p, col);
    if (deduct < infty) p->lb += deduct;
  }
}                                                                  

touredge *pick_edge (problem *p) {
  touredge *cand;
  int cand_inc = -1;
  int testrow, testcol, i, alpha, beta, this_inc;
  problem test;
 
  cand = (touredge *) malloc (sizeof(touredge));
  for (testrow = N-1; testrow>=0; testrow--) {
    for (testcol = N-1; testcol>=0; testcol--) {
        if (p->edge[testrow][testcol]==0) {
          alpha = infty; beta = infty;
          for (i=0; i<N; i++) {
            if ((alpha > p->edge[testrow][i]) && (i!=testcol))
              alpha = p->edge[testrow][i];
            if ((beta > p->edge[i][testcol]) && (i!=testrow))
              beta = p->edge[i][testcol];
          }
          if ((alpha == infty) && (beta == infty)) {
             this_inc = -2;
          } else if (alpha == infty) {
            this_inc = beta;
          } else if (beta == infty) {
            this_inc = alpha;
          } else this_inc = alpha + beta;
 
          if (this_inc >= cand_inc) {
            cand_inc  = this_inc;
            cand->from = testrow;
            cand->to   = testcol;
        }
        cand->from = testrow;
        cand->to   = testcol;
        return cand;
      }
    }
  }
  /* if (cand_inc == -1) {
    return NULL;
  }
  else return cand;  */
  return NULL;
}                                                                

void show_tour (problem *cand)
{
  int i;
 
  printf("Tour has length %d\n\n Edges are ", cand->lb);
  for (i=0; i<N; i++) {
    printf("(%d, %d) ",cand->solution.tour[i].from,cand->solution.tour[i].to);
  }
  printf("\n\n");
}

void copy_problem (problem *to, problem *from)
{
  int i, r, c;

  for (r=0; r<N; r++) {
    for (c=0; c<N; c++) {
      to->edge[r][c] = from->edge[r][c];
    }
  }
  to->lb = from->lb;

  for (i = 0; i<N; i++) to->solution.tour[i] = from->solution.tour[i];
  to->solution.length = from->solution.length;
}

problem *useit (touredge *e, problem *p)
{
  int i;
  problem *newp;
 
  newp = (problem *) malloc (sizeof(problem));
  copy_problem (newp, p);
  newp->solution.tour[newp->solution.length++] = *e;
  for (i=0; i<N; i++) {
     newp->edge[e->from][i] = infty;
     newp->edge[i][e->to] = infty;
  }
  newp->edge[e->to][e->from] = infty;
  newp->edge[e->from][e->to] = infty;
  reduce_matrix(newp);
  return newp;
}
 
problem *binit (touredge *e, problem *p)
{
  problem *newp;
 
  newp = (problem *) malloc (sizeof(problem));
  copy_problem (newp, p);
  newp->edge[e->from][e->to] = infty;
  newp->lb += reduce_row(newp, e->from);
  newp->lb += reduce_col(newp, e->to);
  return newp;
}
                                                                                                       
int is_tour (touredge *cand)
/* assumes cand is a list of N genuine edges */
{
  int enters[N], leaves[N], succ[N];
  int i, next;
 
  for (i=0; i<N; i++) {
    succ[cand[i].from] = cand[i].to;
    leaves[i] = 0;
    enters[i] = 0;
  }
 
  for (i=0, next=0; i<N; i++) {
    leaves[next] = 1;
    next = succ[next];
    enters[next] = 1;
  }
 
  for (i=0; i<N; i++) {
    if (enters[i] == 0) return 0;
    if (leaves[i] == 0) return 0;
  }
  return next==0;
}
                                                   

void show_problem (problem *p)
{
  int i , r, c;

  cout << "----------------------\n\n";

  for (r=0; r<N; r++) {
    for (c=0; c<N; c++) {
      if (p->edge[r][c] < infty) cout << p->edge[r][c] << "  ";
      else cout << "  -";
    }
    cout << "\n";
  }
  cout << "\n LB is " << p->lb << "\n";
  cout << "Tour edges are  ";
  for (i=0; i<p->solution.length; i++) {
   cout << p->solution.tour[i].from << "-" << p->solution.tour[i].to;
  }
  cout << "\n\n----------------------\n";
}


problem *getproblem(Empty dummy)
{
  int r, c;
  problem *p;

  ifstream infile("data.txt");

  p = (problem *) malloc (sizeof(problem));
  for (r=0; r<N; r++)
    for (c=0; c<N; c++) {
      infile >> p->edge[r][c];
      if (p->edge[r][c] == -1) p->edge[r][c] = infty; 
    }
  p->lb = 0;
  p->solution.length = 0;
  reduce_matrix (p);
  return p;
}

bool betterthan(problem x, problem y)
{
  return x.lb < y.lb ;
}

bool isSolution (problem x)
{
  return (x.solution.length == N) && (is_tour(x.solution.tour));
}

void generatecases (Empty dummy){
  problem *cand, *with, *without;
  touredge *key_edge;

  do{
    cand = get<problem>();
    if (cand->solution.length < N)
      key_edge = pick_edge(cand);
      if (key_edge) {
        with     = useit (key_edge, cand);
        put<problem>(*with);
        without  = binit (key_edge, cand);
        put<problem>(*without);
      }
  } while (cand != NULL);}

void display_results(problem cand)
{
  if (is_tour(cand.solution.tour)) show_tour(&cand);
  else cout << "No tour found\n";
}

int main(int argc, char **argv){

// *************************************************************
// Test: test of branch and bound skeleton
// ************************************************************* 

  try{
    InitSkeletons(argc,argv);
// step 1: create a process topology (using C++ constructors)
    Initial<problem>             p1(getproblem); 
    Filter<problem, problem>     p2(generatecases,1);
    BranchAndBound<problem>      p3(p2,2,betterthan,isSolution); 
    Final<problem>               p4(& display_results);  
    Pipe                         p5(p1,p3);
    Pipe                         p6(p5,p4);

    // if (Debug) p6.show();

// step 2: 
// (each processor will pick its part of the computation
// depending on its rank)
    
    p6.start();

    TerminateSkeletons();}
  catch(Exception&){cout << "Exception" << endl << flush;}
}

