#include "Skeleton.h"

inline double init(const int a, const int b){return (a==b) ? 1.0 : 2.0;}

inline double copyPivot(const DistributedMatrix<double>& A, int k, int i, int j, double Pij){
  return A.isLocal(k,k) ? A.get(k,j)/A.get(k,k) : 0;}

inline void pivotOp(const DistributedMatrix<double>& Pivot, int rows, int firstrow,
            int k, double** A){
  double Alk;
  for (int l=0; l<rows; l++){
    Alk = A[l][k];
    for (int j=k; j<=Problemsize; j++)
      if (firstrow+l == k)
        A[l][j] = Pivot.getLocalGlobal(0,j);
      else  A[l][j] -=  Alk * Pivot.getLocalGlobal(0,j);}}

// idea: - each processor gets a sequence of rows of matrix A
//       - repeatedly the pivot row is broadcasted to everybody
//         and each processor executes the pivot operation on its partition of A
void gauss(DistributedMatrix<double>& A){
  // for simplicity assuming pivot != 0
  DistributedMatrix<double> Pivot(sk_numprocs,Problemsize+1,0.0,sk_numprocs,1);
  for (int k=0; k<Problemsize; k++){
    Pivot.mapIndexInPlace(curry(copyPivot)(A)(k));
    Pivot.broadcastPartition(k/A.getLocalRows(),0); 
    A.mapPartitionInPlace(curry(pivotOp)(Pivot,A.getLocalRows(),A.getFirstRow(),k));}} 

int main(int argc, char **argv){
//*************************************************************
// Test: Gaussian Elimination
// *************************************************************
  try{
    InitSkeletons(argc,argv);
   
    DistributedMatrix<double> A(Problemsize,Problemsize+1,& init,sk_numprocs,1);

#if Output == 1
    if (sk_myid == 0) {cout << "A:" << endl << flush;} 
    A.show();
#endif

    gauss(A);

#if Output == 1
    if (sk_myid == 0) {cout << "Result of Gaussian Elimination:" << endl << flush;}
    A.show();
#endif

    TerminateSkeletons();}
  catch(Exception&){cout << "Exception" << endl << flush;}               
}
