#include "Skeleton.h"
#include "math.h"

inline int negate(const int a){return -a;}

template <class C>
inline int copy(const DistributedMatrix<C>& A, int row, int col, C x){return A.get(row,col);}

inline int min(const int a, const int b){return a<b ? a : b;}

const int infinity = 999999;

inline int init(int x, int y){return abs(x-y) <= 1 ? abs(x-y) : infinity;}

template <class C> // (min,+) skalar product
C skprod(const DistributedMatrix<C>& A,const DistributedMatrix<C>& B, int i, int j, C Cij){
  C sum = Cij;
  for (int k=0; k<A.getLocalRows(); k++)
     sum = min(sum, A.getGlobalLocal(i,k) + B.getLocalGlobal(k,j));
  return sum;}

template <class C> // using algorithm of Gentleman based on torus topology
DistributedMatrix<C> matmult(DistributedMatrix<C>& A, DistributedMatrix<C>& B){
    // assumption: A and B are DIFFERENT matrices and have the same square shape
    A.rotateRows(& negate);
    B.rotateCols(& negate); 
    DistributedMatrix<C> R(A.getRows(), A.getCols(),infinity, A.getBlocksInCol(), A.getBlocksInRow());
    for (int i=0; i< A.getBlocksInRow(); i++){  
      typedef C (*skprod_t)(const DistributedMatrix<C>&, 
                            const DistributedMatrix<C>&, int, int, C);
      R.mapIndexInPlace(curry((skprod_t)skprod<int>)(A)(B)); 
      A.rotateRows(-1);
      B.rotateCols(-1);}
    return R;}

template <class C> // assumption: Problemsize is power of 2
DistributedMatrix<C> shortestPaths(DistributedMatrix<C>& A){
  DistributedMatrix<C> B(A.getRows(), A.getCols(),0,A.getBlocksInCol(), A.getBlocksInRow());
  for(int i=0; i<log2(Problemsize); i++){
    typedef int (*copy_t)( const DistributedMatrix<C>&,int,int,C);
    B.mapIndexInPlace(curry((copy_t)copy<C>)(A)); // B is copy of A 
    A = matmult(A,B);}
  return A;}

int main(int argc, char **argv){
//*************************************************************
// Test: all pairs of shortest paths
// *************************************************************
  try{
    InitSkeletons(argc,argv);

    int sqrtp = (int) (sqrt(sk_numprocs)+0.1);
    DistributedMatrix<int> A(Problemsize,Problemsize,init,sqrtp,sqrtp);

#if Output == 1
    if (sk_myid == 0) {cout << "A:" << endl << flush;} 
    A.show();
#endif

    DistributedMatrix<int> B =  shortestPaths(A);

#if Output == 1
    if (sk_myid == 0) {cout << "Result:" << endl << flush;} 
    B.show();
#endif

    TerminateSkeletons();}
  catch(Exception&){cout << "Exception" << endl << flush;} 
}
